from .audit import Audit
from .audit_printer import AuditPrinter
from .audit_referencial_integrity import AuditReferentialIntegrity
from .audit_ttls import AuditTimeToLiveConfigurations

__all__ = (
    "Audit",
    "AuditPrinter",
    "AuditTimeToLiveConfigurations",
    "AuditReferentialIntegrity",
)



import asyncio
from dataclasses import asdict
from logging import getLogger

from ..model import (
    IngestionHookRunRequest,
    Node,
    NodeCreationRule,
    RelationshipWithNodes,
    TimeToLiveConfiguration,
)
from ..schema.indexes import FieldIndex, KeyIndex
from .ingest_strategy import IngestionStrategy
from .operation_debouncer import OperationDebouncer
from .query_executor import QueryExecutor


class DebouncedIngestStrategy(IngestionStrategy, alias="debounced"):
    def __init__(self, query_executor: QueryExecutor) -> None:
        self.logger = getLogger(self.__class__.__name__)
        self.executor = query_executor
        self.debouncer = OperationDebouncer()
        self.hooks_saved_for_after_ingest = []

    async def ingest_source_node(self, source: Node):
        self.debouncer.debounce_node_operation(
            source, node_creation_rule=NodeCreationRule.EAGER
        )

    async def ingest_relationship(self, relationship: RelationshipWithNodes):
        self.debouncer.debounce_node_operation(
            relationship.from_node, relationship.from_side_node_creation_rule
        )
        self.debouncer.debounce_node_operation(
            relationship.to_node, relationship.to_side_node_creation_rule
        )
        self.debouncer.debounce_relationship(relationship)

    async def run_hook(self, request: IngestionHookRunRequest):
        if request.before_ingest:
            await self.executor.execute_hook(request.hook)
        else:
            self.hooks_saved_for_after_ingest.append(request.hook)

    async def upsert_key_index(self, index: KeyIndex):
        self.logger.debug("Ensuring Index Created", extra=asdict(index))
        await self.executor.upsert_key_index(index)
        self.logger.info("Ensured Index Created", extra=asdict(index))

    async def upsert_field_index(self, index: FieldIndex):
        self.logger.debug("Ensuring Index Created", extra=asdict(index))
        await self.executor.upsert_field_index(index)
        self.logger.info("Ensured Index Created", extra=asdict(index))

    async def perform_ttl_operation(self, config: TimeToLiveConfiguration):
        self.logger.debug("Executing TTL", extra=asdict(config))
        await self.executor.perform_ttl_op(config)
        self.logger.info("Executed TTL", extra=asdict(config))

    def flush_nodes_updates(self):
        update_coroutines = (
            self.executor.upsert_nodes_in_bulk_with_same_operation(
                operation, node_group
            )
            for operation, node_group in self.debouncer.drain_node_groups()
        )
        return asyncio.gather(*update_coroutines)

    def flush_relationship_updates(self):
        update_coroutines = (
            self.executor.upsert_relationships_in_bulk_of_same_operation(
                rel_shape, rel_group
            )
            for rel_shape, rel_group in self.debouncer.drain_relationship_groups()
        )
        return asyncio.gather(*update_coroutines)

    async def flush(self):
        await self.flush_nodes_updates()
        await self.flush_relationship_updates()

        # Because we don't know what exactly the hooks do, we can't reliably parallelize
        # them because we could overwhelm the database so we are going to do them one at
        # a time.
        for hook in self.hooks_saved_for_after_ingest:
            await self.executor.execute_hook(hook)
        self.hooks_saved_for_after_ingest.clear()



from pathlib import Path
from typing import Optional

from ...project import PipelineDefinition, PipelineScope, Project
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

DEFAULT_SCOPE_NAME = "default"


class AddPipelineToProject(Operation):
    def __init__(
        self, project: Project, pipeline_path: Path, scope_name: Optional[str] = None
    ) -> None:
        self.project = project
        self.pipeline_path = pipeline_path
        self.scope_name = scope_name or DEFAULT_SCOPE_NAME

    async def perform(self, _: NodestreamCommand):
        scope = self.get_scope()
        scope.add_pipeline_definition(PipelineDefinition.from_path(self.pipeline_path))

    def get_scope(self) -> PipelineScope:
        if self.scope_name in self.project.scopes_by_name:
            return self.project.scopes_by_name[self.scope_name]

        scope = PipelineScope(self.scope_name, {})
        self.project.add_scope(scope)
        return scope



from dataclasses import dataclass
from typing import Any, Dict, List

UNWIND_QUERY = "UNWIND $batched_parameter_sets as params RETURN params"
COMMIT_QUERY = """
CALL apoc.periodic.iterate(
    $iterable_query,
    $batched_query,
    {batchsize: 1000, parallel: true, retries: 3, params: $iterate_params}
)
YIELD batches, committedOperations, failedOperations, errorMessages
RETURN batches, committedOperations, failedOperations, errorMessages
"""

NON_APOCH_COMMIT_QUERY = """
UNWIND $iterate_params.batched_parameter_sets AS param
CALL apoc.cypher.doIt($batched_query, {params: param})
YIELD value
RETURN value
"""


@dataclass(slots=True, frozen=True)
class Query:
    query_statement: str
    parameters: Dict[str, Any]

    @classmethod
    def from_statement(cls, statement: str):
        return cls(query_statement=statement, parameters={})

    def feed_batched_query(self, batched_query: str) -> "Query":
        """Feed the results of the the query into another query that will be executed in batches."""
        return Query(
            COMMIT_QUERY,
            {
                "iterate_params": self.parameters,
                "batched_query": batched_query,
                "iterable_query": self.query_statement,
            },
        )


@dataclass(slots=True, frozen=True)
class QueryBatch:
    query_statement: str
    batched_parameter_sets: List[Dict[str, Any]]

    def as_query(self, apoc_iterate: bool) -> Query:
        return Query(
            {True: COMMIT_QUERY, False: NON_APOCH_COMMIT_QUERY}[apoc_iterate],
            {
                "iterate_params": {
                    "batched_parameter_sets": self.batched_parameter_sets
                },
                "batched_query": self.query_statement,
                "iterable_query": UNWIND_QUERY,
            },
        )



from cleo.helpers import option

from ..operations import InitializeLogger, InitializeProject, RunPipeline
from .nodestream_command import NodestreamCommand
from .shared_options import JSON_OPTION, MANY_PIPELINES_ARGUMENT, PROJECT_FILE_OPTION


class Run(NodestreamCommand):
    name = "run"
    description = "run a pipeline in the current project"
    arguments = [MANY_PIPELINES_ARGUMENT]
    options = [
        PROJECT_FILE_OPTION,
        JSON_OPTION,
        option(
            "annotations",
            "a",
            "An annotation to apply to the pipeline during initialization. Steps without one of these annotations will be skipped.",
            multiple=True,
            flag=False,
        ),
        option(
            "reporting-frequency",
            "r",
            "How often to report progress",
            default=1000,
            flag=False,
        ),
        option(
            "step-outbox-size",
            "s",
            "How many records to buffer in each step's outbox before blocking",
            default=1000,
            flag=False,
        ),
        option(
            "target",
            "t",
            "Specify a database to target at run time.",
            multiple=True,
            flag=False,
        ),
    ]

    async def handle_async(self):
        await self.run_operation(InitializeLogger())
        project = await self.run_operation(InitializeProject())
        await self.run_operation(RunPipeline(project))



from typing import Any

from .normalizer import Normalizer


class TrimWhitespace(Normalizer, alias="trim_whitespace"):
    def normalize_value(self, value: Any) -> Any:
        return value.strip() if isinstance(value, str) else value



from dataclasses import dataclass, field
from importlib import resources
from typing import Dict, Optional

from schema import Or

from ..file_io import LoadsFromYamlFile
from ..pipeline.scope_config import ScopeConfig
from .pipeline_definition import PipelineConfiguration, PipelineDefinition
from .pipeline_scope import PipelineScope


@dataclass
class PluginConfiguration(LoadsFromYamlFile):
    """A `PluginConfiguration` represents a collection of configuration for a plugin.

    A config is a key value pair object in a nodestream scope including plugins.
    It contains a collection of configuration key value pairs to be used by the pipelines of a scope.
    """

    name: str
    pipelines_by_name: Dict[str, PipelineDefinition] = field(default_factory=dict)
    config: Optional[ScopeConfig] = None
    pipeline_configuration: PipelineConfiguration = None

    @classmethod
    def describe_yaml_schema(cls):
        from schema import And, Optional, Schema, Use

        return Schema(
            {
                "name": str,
                Optional("config"): ScopeConfig.describe_yaml_schema(),
                Optional("targets"): And(Use(set), {str}),
                Optional("annotations"): {str: Or(str, int, float, bool)},
                Optional("pipelines"): [PipelineDefinition.describe_yaml_schema()],
            }
        )

    @classmethod
    def from_file_data(cls, data) -> "PluginConfiguration":
        name = data.pop("name")
        config = data.pop("config")
        pipelines_data = data.pop("pipelines", [])
        configuration = PipelineConfiguration.from_file_data(data)
        pipelines_by_name = {
            pipeline["name"]: PipelineDefinition.from_plugin_data(
                pipeline, configuration
            )
            for pipeline in pipelines_data
        }
        return cls(
            name,
            pipelines_by_name,
            ScopeConfig.from_file_data(config) if config else None,
            configuration,
        )

    def to_file_data(self):
        data = {"name": self.name}

        if self.config is not None:
            data["config"] = self.config.to_file_data()

        if annotations := self.pipeline_configuration.annotations:
            data["annotations"] = annotations

        if targets := self.pipeline_configuration.targets:
            data["targets"] = list(targets)

        if self.pipelines_by_name:
            data["pipelines"] = [
                ppl.to_plugin_file_data() for ppl in self.pipelines_by_name.values()
            ]

        return data

    def update_configurations(self, other: "PluginConfiguration"):
        """Updates the `PluginConfiguration` using the
        PluginConfiguration from the provided PluginConfiguration object.

        Used for adding project plugin configurations to plugins pipelines loaded from resources.
        """
        self.config = other.config
        for name, pipeline in self.pipelines_by_name.items():
            # set the plugin configuration level targets and annotations
            pipeline.configuration.parent = other.pipeline_configuration

            # override the pipeline configuration if they are available
            other_pipeline = other.pipelines_by_name.get(name)
            if other_pipeline is not None:
                pipeline.use_configuration(other_pipeline.configuration)

    def make_scope(self) -> PipelineScope:
        """Creates a `PipelineScope` object from the `PluginConfiguration` object"""
        return PipelineScope(self.name, self.pipelines_by_name, False, self.config)

    @classmethod
    def from_resources(
        cls, name: str, package: resources.Package
    ) -> "PluginConfiguration":
        """Load a `PluginConfiguration` from a package's resources.

        Each `.yaml` file in the package's resources will be loaded as a pipeline.
        Internally, this uses `importlib.resources` to load the files and calls
        `PipelineDefinition.from_path` on each file. This means that the name of
        the pipeline will be the name of the file without the .yaml extension.

        Args:
            name: The name of the scope.
            package: The name of the package to load from.

        Returns:
            A `PluginConfiguration` instance with the pipelines defined in the package.
        """
        pipelines = [
            PipelineDefinition.from_path(f)
            for f in resources.files(package).iterdir()
            if f.suffix == ".yaml"
        ]
        pipelines_by_name = {pipeline.name: pipeline for pipeline in pipelines}
        return cls(name, pipelines_by_name)



import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

from cleo.commands.command import Command
from cleo.io.outputs.output import Verbosity

from ...pluggable import Pluggable
from ...project import Project

if TYPE_CHECKING:
    from ..operations import Operation

DEFAULT_PROJECT_FILE = Path("nodestream.yaml")


class NodestreamCommand(Command, Pluggable):
    entrypoint_name = "commands"

    def handle(self):
        return asyncio.run(self.handle_async())

    async def handle_async(self):
        raise NotImplementedError

    async def run_operation(self, operation: "Operation"):
        self.line(
            f"<info>Running: {operation.name}</info>", verbosity=Verbosity.VERBOSE
        )
        return await operation.perform(self)

    def get_project_path(self) -> Path:
        path = self.option("project")
        return DEFAULT_PROJECT_FILE if path is None else Path(path)

    def get_project(self) -> Project:
        return Project.read_from_file(self.get_project_path())

    @property
    def has_json_logging_set(self) -> bool:
        return self.option("json")

    @property
    def scope(self) -> str:
        return self.option("scope")

    @property
    def is_verbose(self) -> bool:
        return self.io.output.is_verbose()

    @property
    def is_very_verbose(self) -> bool:
        return self.io.output.is_very_verbose()



import re
from typing import Any, Iterable, Type

from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import ValueProvider


class RegexValueProvider(ValueProvider):
    def __init__(self, regex: str, data: ValueProvider, group: int | str = 0):
        self.regex = re.compile(regex)
        self.group = group
        self.data = data

    def apply_regex_to_value(self, value: Any) -> Any:
        match = self.regex.match(value)
        if match:
            return match.group(self.group)
        return None

    def single_value(self, context: ProviderContext) -> Any:
        return self.apply_regex_to_value(self.data.single_value(context))

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        return map(self.apply_regex_to_value, self.data.many_values(context))

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!regex", lambda loader, node: cls(**loader.construct_mapping(node))
        )


SafeDumper.add_representer(
    RegexValueProvider,
    lambda dumper, regex: dumper.represent_mapping(
        "!regex",
        {"regex": regex.regex.pattern, "group": regex.group, "data": regex.data},
    ),
)



from collections import defaultdict
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import Any, Dict

UNKNOWN_PIPELINE_NAME = "unknown"


@dataclass(frozen=True, slots=True)
class PipelineContext:
    name: str = UNKNOWN_PIPELINE_NAME
    stats: Dict[str, Any] = field(default_factory=lambda: defaultdict(int))

    def increment_stat(self, stat_name: str, amount: int = 1):
        self.stats[stat_name] += amount


context: ContextVar[PipelineContext] = ContextVar("context")


def get_context() -> PipelineContext:
    try:
        return context.get()
    except LookupError:
        return PipelineContext()


@contextmanager
def start_context(pipeline_name: str):
    token = context.set(PipelineContext(pipeline_name))
    try:
        yield
    finally:
        context.reset(token)



from ..pipeline import Writer
from .database_connector import DatabaseConnector
from .debounced_ingest_strategy import DebouncedIngestStrategy
from .ingest_strategy import INGESTION_STRATEGY_REGISTRY, IngestionStrategy


class GraphDatabaseWriter(Writer):
    @classmethod
    def from_file_data(
        cls,
        database: str,
        ingest_strategy_name: str = INGESTION_STRATEGY_REGISTRY.name_for(
            DebouncedIngestStrategy
        ),
        collect_stats: bool = True,
        batch_size: int = 1000,
        **database_args
    ):
        connector = DatabaseConnector.from_database_args(
            database=database, **database_args
        )
        executor = connector.get_query_executor(collect_stats=collect_stats)
        ingest_strategy_cls = INGESTION_STRATEGY_REGISTRY.get(ingest_strategy_name)
        ingest_strategy = ingest_strategy_cls(executor)

        return cls(
            batch_size=batch_size,
            ingest_strategy=ingest_strategy,
        )

    def __init__(self, batch_size: int, ingest_strategy: IngestionStrategy) -> None:
        self.batch_size = batch_size
        self.ingest_strategy = ingest_strategy
        self.pending_records = 0

    async def flush(self):
        await self.ingest_strategy.flush()
        self.pending_records = 0

    async def write_record(self, ingestible):
        await ingestible.ingest(self.ingest_strategy)
        self.pending_records += 1
        if self.pending_records >= self.batch_size:
            await self.flush()



from typing import Any, Dict, Iterable, Optional

from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...schema.schema import (
    GraphObjectShape,
    GraphObjectType,
    PropertyMetadataSet,
    UnknownTypeMarker,
)
from .interpretation import Interpretation


class PropertiesInterpretation(Interpretation, alias="properties"):
    """Stores additional properties onto the source node."""

    __slots__ = ("properties", "norm_args")

    def __init__(
        self,
        properties: StaticValueOrValueProvider,
        normalization: Optional[Dict[str, Any]] = None,
    ):
        self.properties = ValueProvider.guarantee_provider_dictionary(properties)
        self.norm_args = normalization or {}

    def interpret(self, context: ProviderContext):
        source = context.desired_ingest.source
        source.properties.apply_providers(context, self.properties, self.norm_args)

    def gather_object_shapes(self) -> Iterable[GraphObjectShape]:
        yield GraphObjectShape(
            graph_object_type=GraphObjectType.NODE,
            object_type=UnknownTypeMarker.source_node(),
            properties=PropertyMetadataSet.from_names(self.properties.keys()),
        )



from .extract_variables_interpretation import ExtractVariablesInterpretation
from .interpretation import Interpretation
from .properties_interpretation import PropertiesInterpretation
from .relationship_interpretation import RelationshipInterpretation
from .source_node_interpretation import SourceNodeInterpretation
from .switch_interpretation import SwitchInterpretation

__all__ = (
    "ExtractVariablesInterpretation",
    "Interpretation",
    "PropertiesInterpretation",
    "RelationshipInterpretation",
    "SourceNodeInterpretation",
    "SwitchInterpretation",
)



from typing import Any, Iterable, Type

from yaml import SafeLoader

from ..normalizers import Normalizer
from .context import ProviderContext
from .value_provider import ValueProvider


class NormalizerValueProvider(ValueProvider):
    def __init__(self, using: str, data: ValueProvider):
        self.normalizer = Normalizer.from_alias(using)
        self.data = data

    def single_value(self, context: ProviderContext) -> Any:
        return next(self.many_values(context))

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        return map(self.normalizer.normalize_value, self.data.many_values(context))

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!normalize", lambda loader, node: cls(**loader.construct_mapping(node))
        )



from ..operations import InitializeProject, ShowPipelines
from .nodestream_command import NodestreamCommand
from .shared_options import JSON_OPTION, PROJECT_FILE_OPTION, SCOPE_NAME_OPTION


class Show(NodestreamCommand):
    name = "show"
    description = "Show the Pipelines in your Nodestream Project"
    options = [PROJECT_FILE_OPTION, SCOPE_NAME_OPTION, JSON_OPTION]

    async def handle_async(self):
        project = await self.run_operation(InitializeProject())
        use_json = self.has_json_logging_set
        await self.run_operation(ShowPipelines(project, self.scope, use_json))



from typing import Any, Dict, Iterable

from ...model import TimeToLiveConfiguration
from ...schema.schema import GraphObjectType
from .extractor import Extractor


class TimeToLiveConfigurationExtractor(Extractor):
    def __init__(
        self, graph_object_type: str, configurations: Iterable[Dict[str, Any]]
    ) -> None:
        self.graph_object_type = GraphObjectType(graph_object_type)
        self.configurations = configurations

    async def extract_records(self):
        for config in self.configurations:
            yield TimeToLiveConfiguration(
                **config, graph_object_type=self.graph_object_type
            )



from io import StringIO
from logging import getLogger
from pathlib import Path
from typing import Any, AsyncGenerator, Optional

from ...extractor import Extractor
from ...files import SupportedFileFormat
from .credential_utils import AwsClientFactory


class S3Extractor(Extractor):
    @classmethod
    def from_file_data(
        cls,
        bucket: str,
        prefix: Optional[str] = None,
        archive_dir: Optional[str] = None,
        object_format: Optional[str] = None,
        **aws_client_args,
    ):
        return cls(
            bucket=bucket,
            object_format=object_format,
            prefix=prefix,
            archive_dir=archive_dir,
            s3_client=AwsClientFactory(**aws_client_args).make_client("s3"),
        )

    def __init__(
        self,
        bucket: str,
        s3_client,
        archive_dir: Optional[str] = None,
        object_format: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> None:
        self.object_format = object_format
        self.prefix = prefix or ""
        self.bucket = bucket
        self.archive_dir = archive_dir
        self.s3_client = s3_client
        self.logger = getLogger(__name__)

    def get_object_as_io(self, key: str) -> StringIO:
        return self.s3_client.get_object(Bucket=self.bucket, Key=key)["Body"]

    def archive_s3_object(self, key: str):
        if self.archive_dir:
            self.logger.info("Archiving S3 Object", extra=dict(key=key))
            filename = Path(key).name
            self.s3_client.copy_object(
                Bucket=self.bucket,
                Key=f"{self.archive_dir}/{filename}",
                CopySource={"Bucket": self.bucket, "Key": key},
            )
            self.s3_client.delete_object(Bucket=self.bucket, Key=key)

    def infer_object_format(self, key: str) -> str:
        object_format = self.object_format or Path(key).suffix
        if not object_format:
            raise ValueError(
                f"No object format provided and key has no extension: '{key}'"
            )
        return object_format

    def get_object_as_file(self, key: str) -> SupportedFileFormat:
        io = self.get_object_as_io(key)
        object_format = self.infer_object_format(key)
        return SupportedFileFormat.from_file_pointer_and_format(io, object_format)

    def is_object_in_archive(self, key: str) -> bool:
        if self.archive_dir:
            return key.startswith(self.archive_dir)
        return False

    def find_keys_in_bucket(self) -> list[str]:
        # Returns all keys in the bucket that are not in the archive dir and have the prefix.
        paginator = self.s3_client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=self.bucket, Prefix=self.prefix)
        for page in page_iterator:
            keys = (obj["Key"] for obj in page.get("Contents", []))
            yield from filter(lambda k: not self.is_object_in_archive(k), keys)

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for key in self.find_keys_in_bucket():
            try:
                for record in self.get_object_as_file(key).read_file():
                    yield record
            finally:
                self.archive_s3_object(key)



from dataclasses import dataclass, field
from typing import Any, Dict

from ...model import DesiredIngestion, JsonLikeDocument, PropertySet


@dataclass(slots=True)
class ProviderContext:
    """Defines the state of the Interpretation at a given point of time.

    As a record from the pipeline traverses the series of components responsible for extracting data from it and
    converting it into a SubGraph, data must be viewed within the context of that pipeline. Variables and data mappings
    are stored that are later referenced by disparate components, each of which, coordinate through an `ProviderContext`.

    This data does live longer then the record's time spent in the pipeline and is discarded excepted for the stored `DesiredIngestion`.
    The `DesiredIngestion` represents the outcome of the work spent on the record as it passes through the pipeline.
    """

    document: JsonLikeDocument
    desired_ingest: DesiredIngestion
    mappings: Dict[Any, Any] = field(default_factory=dict)
    variables: PropertySet = field(default_factory=PropertySet.empty)

    @classmethod
    def fresh(cls, record):
        return cls(record, DesiredIngestion())



from yaml import load

from ...file_io import LazyLoadedArgument
from .argument_resolver import ArgumentResolver


class IncludeFileResolver(ArgumentResolver, alias="include"):
    """An `IncludeFileResolver` is an `ArgumentResolver` that can resolve a file path into a file's contents."""

    @staticmethod
    def resolve_argument(file_path: str):
        from ..pipeline_file_loader import PipelineFileContents

        with open(file_path) as f:
            data = load(f, Loader=PipelineFileContents.get_loader())
            return LazyLoadedArgument.resolve_if_needed(data)



from typing import Any, AsyncGenerator, Iterable

from .extractor import Extractor


class IterableExtractor(Extractor):
    """An extractor that produces records from an iterable."""

    @classmethod
    def range(cls, start=0, stop=100, step=1):
        return cls(iterable=({"index": i} for i in range(start, stop, step)))

    def __init__(self, iterable: Iterable[Any]) -> None:
        self.iterable = iterable

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for record in self.iterable:
            yield record



from typing import Any, Dict, Optional

from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from .interpretation import Interpretation


class ExtractVariablesInterpretation(Interpretation, alias="variables"):
    """Stores variables that can be used later in the processing of a record.

    You may store an arbitrary set of properties as variables that come from any value provider or statically provided.

    ```yaml
    interpretations:
      - type: variables
        variables:
          first_name: !jmespath first_name
          last_name: Smith
    ```

    You may also apply normalization in the same way as any other interpretation.

    ```yaml
    interpretations:
      - type: variables
        variables:
          first_name: !jmespath first_name
        normalization:
            do_lowercase_strings: true
    ```
    """

    __slots__ = ("variables", "norm_args")

    def __init__(
        self,
        variables: Dict[str, StaticValueOrValueProvider],
        normalization: Optional[Dict[str, Any]] = None,
    ):
        self.variables = ValueProvider.guarantee_provider_dictionary(variables)
        self.norm_args = normalization or {}

    def interpret(self, context: ProviderContext):
        context.variables.apply_providers(context, self.variables, self.norm_args)



import asyncio
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from typing import Any, AsyncGenerator, Dict, Optional

from ..class_loader import ClassLoader
from ..flush import Flush
from ..step import Step
from ..value_providers import ProviderContext, StaticValueOrValueProvider, ValueProvider


class Transformer(Step):
    """A `Transformer` takes a given record and mutates into a new record.

    `Transformer` steps generally make up the middle of an ETL pipeline and are responsible
    for reshaping an object so its more ingestible by the downstream sink.
    """

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        async for record in record_stream:
            if record is Flush:
                yield record
            else:
                val_or_gen = self.transform_record(record)
                if isinstance(val_or_gen, AsyncGenerator):
                    async for result in val_or_gen:
                        yield result
                else:
                    yield await val_or_gen

    @abstractmethod
    async def transform_record(self, record: Any) -> Any:
        raise NotImplementedError


class ConcurrentTransformer(Transformer):
    """A `ConcurrentTransformer` takes a given record and mutates into a new record while managing a pool of concurrent workers.

    `ConcurrentTransformer`s are useful when the transformation work is IO bound and can be parallelized across multiple threads.
    `ConcurrentTransformer`s are not useful when the transformation work is CPU bound as the GIL will prevent useful parallelization.
    Additionally, `ConcurrentTransformer`s do not guarantee the order of the output stream will match the input stream.
    """

    def __init__(
        self, thread_pool_size: Optional[int] = None, maximum_pending_tasks: int = 1000
    ) -> None:
        self.logger = getLogger(self.__class__.__name__)
        self.maximum_pending_tasks = maximum_pending_tasks
        self.thread_pool = ThreadPoolExecutor(
            max_workers=thread_pool_size,
            thread_name_prefix=self.__class__.__name__,
            initializer=self.on_worker_thread_start,
        )

    def on_worker_thread_start(self):
        pass

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        pending_tasks = []

        def drain_completed_tasks():
            tasks_drained = 0
            remaining_pending_tasks = []
            for task in pending_tasks:
                if task.done():
                    yield task.result()
                    tasks_drained += 1
                else:
                    remaining_pending_tasks.append(task)

            pending_tasks[:] = remaining_pending_tasks
            if tasks_drained:
                self.logger.debug(
                    "Drained %s completed tasks, %s pending tasks remain",
                    tasks_drained,
                    len(pending_tasks),
                )

        async def drain_completed_tasks_async():
            for result in drain_completed_tasks():
                yield result
            await self.yield_processor()

        async for record in record_stream:
            if record is Flush:
                # Flush the pending tasks, then yield the flush.
                # In order to fully respect the Flush,
                # we need to wait for all pending tasks to complete.
                # If we're still waiting we can yield the processor.
                while pending_tasks:
                    async for result in drain_completed_tasks_async():
                        yield result
                yield record
            else:
                # Submit the work to the thread pool (only if we have capacity)
                # If we don't have capacity, yield completed tasks.
                # Once there is nothing remaining to yield from these tasks, yield the processor until capacity is available.
                submitted = False
                while not submitted:
                    if len(pending_tasks) < self.maximum_pending_tasks:
                        task = self.thread_pool.submit(self.do_work_on_record, record)
                        pending_tasks.append(task)
                        submitted = True
                    async for result in drain_completed_tasks_async():
                        yield result

        # After we've finished enqueuing records, we need to drain all tasks.
        # Items yielded from this loop are the final results of the step.
        self.logger.debug("Finished enqueuing records, draining pending tasks")
        while pending_tasks:
            async for result in drain_completed_tasks_async():
                yield result

    async def finish(self):
        self.thread_pool.shutdown(wait=True)

    async def yield_processor(self):
        await asyncio.sleep(0)

    def do_work_on_record(self, record: Any) -> Any:
        # Handles the work nessary to transform a single record.
        # This method wraps the `transform_record` method in an asyncio loop.
        # This is necessary because the thread pool executor is not async.
        return asyncio.run(self.transform_record(record))

    @abstractmethod
    async def transform_record(self, record: Any) -> Any:
        raise NotImplementedError


class PassTransformer(Transformer):
    async def transform_record(self, record: Any) -> AsyncGenerator:
        yield record


class SwitchTransformer(Transformer):
    __slots__ = (
        "switch_on",
        "transformers",
        "default",
        "normalization",
    )

    @classmethod
    def from_file_data(cls, switch_on, cases, default=None, normalization=None):
        transformers = {
            field_value: ClassLoader().load_class(**transformer)
            for field_value, transformer in cases.items()
        }
        default = ClassLoader().load_class(**default) if default else None
        return cls(switch_on, transformers, default, normalization)

    def __init__(
        self,
        switch_on: StaticValueOrValueProvider,
        transformers: Dict[str, Transformer],
        default: Transformer = None,
        normalization: Dict[str, Any] = None,
    ):
        self.switch_on = ValueProvider.guarantee_value_provider(switch_on)
        self.transformers = transformers
        self.default = default if default else PassTransformer()
        self.normalization = normalization or {}

    async def transform_record(self, record: Any):
        context = ProviderContext.fresh(record)
        key = self.switch_on.normalize_single_value(context, self.normalization)
        transformer = self.transformers.get(key, self.default)
        # If we don't have a transformer to handle the record we just yield the record with PassTransformer
        async for result in transformer.transform_record(record):
            yield result



import json
from typing import List, Union

from ...model import JsonLikeDocument
from .transformer import Transformer


class ExpandJsonField(Transformer):
    @classmethod
    def from_file_data(cls, /, path: Union[str, List[str]]):
        if isinstance(path, str):
            path = [path]

        return cls(path)

    def __init__(self, path: List[str]) -> None:
        self.path = path

    def dig_to_path(self, record: JsonLikeDocument, path: List[str]):
        value_at_path = record
        for path_segment in path:
            value_at_path = value_at_path.get(path_segment, None)
            if value_at_path is None:
                return None

        return value_at_path

    def replace_json_value(self, innermost_dictionary, key):
        json_string = innermost_dictionary.get(key, None)
        if json_string is None:
            return
        innermost_dictionary[key] = json.loads(json_string)

    async def transform_record(self, record: JsonLikeDocument):
        last_segment = self.path[-1]
        innermost_dictionary = self.dig_to_path(record, self.path[:-1])
        if innermost_dictionary is None:
            return record
        self.replace_json_value(innermost_dictionary, last_segment)
        return record



from .apis import SimpleApiExtractor
from .extractor import Extractor
from .files import FileExtractor, RemoteFileExtractor
from .iterable import IterableExtractor
from .ttls import TimeToLiveConfigurationExtractor

__all__ = (
    "Extractor",
    "IterableExtractor",
    "FileExtractor",
    "RemoteFileExtractor",
    "TimeToLiveConfigurationExtractor",
    "SimpleApiExtractor",
)



from typing import Any, Iterable, Type

import jmespath
from jmespath.parser import ParsedResult
from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import ValueProvider


class JmespathValueProvider(ValueProvider):
    """A `ValueProvider` that uses JMESPath to extract values from a document."""

    __slots__ = ("compiled_query",)

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!jmespath",
            lambda loader, node: cls.from_string_expression(
                loader.construct_scalar(node)
            ),
        )

    @classmethod
    def from_string_expression(cls, expression: str):
        return cls(jmespath.compile(expression))

    def __init__(self, compiled_query: ParsedResult) -> None:
        self.compiled_query = compiled_query

    def search(self, context: ProviderContext):
        raw_search = self.compiled_query.search(context.document)
        if raw_search is None:
            return
        if isinstance(raw_search, list):
            yield from raw_search
        else:
            yield raw_search

    def single_value(self, context: ProviderContext) -> Any:
        return next(self.search(context), None)

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        return self.search(context)


SafeDumper.add_representer(
    JmespathValueProvider,
    lambda dumper, jmespath: dumper.represent_scalar(
        "!jmespath", jmespath.compiled_query.expression
    ),
)



from ...project import Project
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class InitializeProject(Operation):
    async def perform(self, command: NodestreamCommand) -> Project:
        return command.get_project()



from pathlib import Path

from ...project import Project
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class CommitProjectToDisk(Operation):
    def __init__(self, project: Project, project_path: Path) -> None:
        self.project = project
        self.project_path = project_path

    async def perform(self, _: NodestreamCommand):
        self.project.write_to_file(self.project_path)



from collections import defaultdict
from typing import Dict, Iterable, Tuple

from ..model.graph_objects import (
    DeduplicatableObject,
    Node,
    NodeCreationRule,
    RelationshipWithNodes,
)
from .query_executor import OperationOnNodeIdentity, OperationOnRelationshipIdentity


class DeduplicationBucket:
    __slots__ = ("bucket",)

    def __init__(self) -> None:
        self.bucket: Dict[tuple, DeduplicatableObject] = {}

    def include(self, deduplicatable: DeduplicatableObject):
        key = deduplicatable.get_dedup_key()
        if existing_operation := self.bucket.get(key):
            existing_operation.update(deduplicatable)
        else:
            self.bucket[key] = deduplicatable

    def values(self) -> Iterable[DeduplicatableObject]:
        return self.bucket.values()


class OperationBucketGroup:
    __slots__ = ("buckets",)

    def __init__(self) -> None:
        self.buckets = defaultdict(DeduplicationBucket)

    def get_bucket(self, key: Tuple) -> DeduplicationBucket:
        return self.buckets[key]

    def drain(self) -> Iterable[Tuple[Tuple, Iterable[DeduplicatableObject]]]:
        for key, bucket in self.buckets.items():
            yield key, bucket.values()

        self.buckets.clear()


class OperationDebouncer:
    __slots__ = ("node_operation_buckets", "relationship_operation_buckets")

    def __init__(self):
        self.node_operation_buckets = OperationBucketGroup()
        self.relationship_operation_buckets = OperationBucketGroup()

    def bucketize_node_operation(
        self, node: Node, node_creation_rule: NodeCreationRule
    ) -> DeduplicationBucket:
        return self.node_operation_buckets.get_bucket(
            OperationOnNodeIdentity(node.identity_shape, node_creation_rule)
        )

    def bucketize_relationship_operation(
        self, relationship: RelationshipWithNodes
    ) -> DeduplicationBucket:
        return self.relationship_operation_buckets.get_bucket(
            OperationOnRelationshipIdentity(
                from_node=OperationOnNodeIdentity(
                    node_identity=relationship.from_node.identity_shape,
                    node_creation_rule=relationship.from_side_node_creation_rule.prevent_creation(),
                ),
                to_node=OperationOnNodeIdentity(
                    node_identity=relationship.to_node.identity_shape,
                    node_creation_rule=relationship.to_side_node_creation_rule.prevent_creation(),
                ),
                relationship_identity=relationship.relationship.identity_shape,
                relationship_creation_rule=relationship.relationship_creation_rule,
            )
        )

    def debounce_node_operation(
        self, node: Node, node_creation_rule: NodeCreationRule = NodeCreationRule.EAGER
    ):
        bucket = self.bucketize_node_operation(node, node_creation_rule)
        bucket.include(node)

    def debounce_relationship(self, relationship: RelationshipWithNodes):
        bucket = self.bucketize_relationship_operation(relationship)
        bucket.include(relationship)

    def drain_node_groups(
        self,
    ) -> Iterable[Tuple[OperationOnNodeIdentity, Iterable[Node]]]:
        yield from self.node_operation_buckets.drain()

    def drain_relationship_groups(
        self,
    ) -> Iterable[
        Tuple[OperationOnRelationshipIdentity, Iterable[RelationshipWithNodes]]
    ]:
        yield from self.relationship_operation_buckets.drain()



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Type, Union

from yaml import SafeLoader

from ...pluggable import Pluggable
from ...subclass_registry import SubclassRegistry
from ..normalizers import Normalizer
from .context import ProviderContext

StaticValueOrValueProvider = Union[Any, "ValueProvider"]


VALUE_PROVIDER_REGISTRY = SubclassRegistry()


@VALUE_PROVIDER_REGISTRY.connect_baseclass
class ValueProvider(Pluggable, ABC):
    """A `ValueProvider` is a class that can extract values from a document."""

    entrypoint_name = "value_providers"

    @classmethod
    def guarantee_value_provider(
        cls, maybe_provider: StaticValueOrValueProvider
    ) -> "ValueProvider":
        from .static_value_provider import StaticValueProvider

        return (
            maybe_provider
            if isinstance(maybe_provider, ValueProvider)
            else StaticValueProvider(maybe_provider)
        )

    @classmethod
    def guarantee_provider_dictionary(
        cls, maybe_providers: Dict[Any, StaticValueOrValueProvider]
    ):
        return {k: cls.guarantee_value_provider(v) for k, v in maybe_providers.items()}

    @classmethod
    def guarantee_provider_list(
        cls, maybe_providers: Iterable[StaticValueOrValueProvider]
    ):
        return [cls.guarantee_value_provider(v) for v in maybe_providers]

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        pass

    @abstractmethod
    def single_value(self, context: ProviderContext) -> Any:
        raise NotImplementedError

    @abstractmethod
    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        raise NotImplementedError

    def normalize(self, value, args):
        return Normalizer.normalize_by_args(value, args)

    def normalize_single_value(
        self, context: ProviderContext, normalization_args
    ) -> Any:
        return self.normalize(self.single_value(context), normalization_args)

    def normalize_many_values(
        self, context: ProviderContext, normalization_args
    ) -> Iterable[Any]:
        for value in self.many_values(context):
            yield self.normalize(value, normalization_args)

    @property
    def is_static(self) -> bool:
        return False



from typing import Any, Iterable

from yaml import SafeDumper

from .context import ProviderContext
from .value_provider import ValueProvider


class StaticValueProvider(ValueProvider):
    """A `ValueProvider` that always returns the same value."""

    __slots__ = ("value",)

    def __init__(self, value) -> None:
        self.value = value

    def single_value(self, _: ProviderContext) -> Any:
        return self.value

    def many_values(self, _: ProviderContext) -> Iterable[Any]:
        return self.value if isinstance(self.value, Iterable) else [self.value]

    @property
    def is_static(self) -> bool:
        return True


SafeDumper.add_representer(
    StaticValueProvider,
    lambda dumper, static: dumper.represent_scalar("!static", static.value),
)



from pathlib import Path
from typing import Optional

from ...project import Project
from ...schema.printers import SchemaPrinter
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class PrintProjectSchema(Operation):
    def __init__(
        self,
        project: Project,
        format_string: str,
        type_overrides_file: Optional[str] = None,
        output_file: Optional[str] = None,
    ) -> None:
        self.project = project
        self.format_string = format_string
        self.output_file = output_file
        self.type_overrides_file = type_overrides_file

    async def perform(self, command: NodestreamCommand):
        type_overrides_file = (
            Path(self.type_overrides_file) if self.type_overrides_file else None
        )
        schema = self.project.get_schema(type_overrides_file=type_overrides_file)

        # Import all schema printers so that they can register themselves
        SchemaPrinter.import_all()
        printer = SchemaPrinter.from_name(self.format_string)
        if self.output_file:
            printer.print_schema_to_file(schema, Path(self.output_file))
        else:
            printer.print_schema_to_stdout(schema, print_fn=command.write)



from dataclasses import dataclass

from ..pipeline import PipelineInitializationArguments
from ..pipeline.meta import start_context
from ..pipeline.progress_reporter import PipelineProgressReporter
from ..pipeline.scope_config import ScopeConfig
from .pipeline_definition import PipelineDefinition


@dataclass
class RunRequest:
    """A `RunRequest` represents a request to run a pipeline.

    `RunRequest` objects should be submitted to a `Project` object to be executed
    via its `run_request` method. The `run_request` method will execute the run request
    on the appopriate pipeline if it exists. Otherwise, it will be a no-op.
    """

    pipeline_name: str
    initialization_arguments: PipelineInitializationArguments
    progress_reporter: PipelineProgressReporter

    @classmethod
    def for_testing(cls, pipeline_name: str, results_list: list) -> "RunRequest":
        """Create a `RunRequest` for testing.

        This method is intended to be used for testing purposes only. It will create a
        run request with the given pipeline name and `PipelineInitializationArguments`
        for testing.

        Args:
            pipeline_name: The name of the pipeline to run.
            results_list: The list to append results to.

        Returns:
            RunRequest: A `RunRequest` for testing.
        """
        return cls(
            pipeline_name,
            PipelineInitializationArguments.for_testing(),
            PipelineProgressReporter.for_testing(results_list),
        )

    async def execute_with_definition(self, definition: PipelineDefinition):
        """Execute this run request with the given pipeline definition.

        This method is intended to be called by `PipelineScope` and should not be called
        directly without good reason. It will execute the run request asynchronously
        with the given pipeline definition. The run request will be executed within a
        context manager that sets the current pipeline name to the name of the pipeline
        being executed.

        Args:
            definition: The pipeline definition to execute this run request with.
        """
        with start_context(self.pipeline_name):
            pipeline = definition.initialize(self.initialization_arguments)
            await pipeline.run(self.progress_reporter)

    def set_configuration(self, scope_config: ScopeConfig):
        self.initialization_arguments.effecitve_config_values = scope_config



from .add_pipeline_to_project import AddPipelineToProject
from .commit_project_to_disk import CommitProjectToDisk
from .generate_pipeline_scaffold import GeneratePipelineScaffold
from .initialize_logger import InitializeLogger
from .initialize_project import InitializeProject
from .operation import Operation
from .print_project_schema import PrintProjectSchema
from .remove_pipeline_from_project import RemovePipelineFromProject
from .run_audit import RunAudit
from .run_copy import RunCopy
from .run_pipeline import RunPipeline
from .run_project_cookiecutter import RunProjectCookiecutter
from .show_pipelines import ShowPipelines

__all__ = (
    "AddPipelineToProject",
    "CommitProjectToDisk",
    "GeneratePipelineScaffold",
    "InitializeLogger",
    "InitializeProject",
    "Operation",
    "RemovePipelineFromProject",
    "RunPipeline",
    "ShowPipelines",
    "PrintProjectSchema",
    "RunProjectCookiecutter",
    "RunAudit",
    "RunCopy",
)



from typing import Dict, Optional

from httpx import AsyncClient

from .extractor import Extractor


class SimpleApiExtractor(Extractor):
    def __init__(
        self,
        url: str,
        yield_from: Optional[str] = None,
        offset_query_param: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self.url = url
        self.offset_query_param = offset_query_param
        self.headers = headers
        self.yield_from = yield_from

    @property
    def does_pagination(self) -> bool:
        return self.offset_query_param is not None

    async def extract_records(self):
        should_continue = True
        records_so_far = 0
        params = None

        async with AsyncClient() as client:
            while should_continue:
                result = await client.get(self.url, headers=self.headers, params=params)
                result.raise_for_status()
                records = result.json()
                if self.yield_from:
                    records = records[self.yield_from]

                for item in records:
                    yield item

                should_continue = len(records) > 0 and self.does_pagination
                records_so_far += len(records)
                params = {self.offset_query_param: records_so_far}



import re
from abc import abstractmethod
from typing import Any, AsyncGenerator, Dict, Iterable, Optional

from .flush import Flush
from .step import Step
from .value_providers import ProviderContext, StaticValueOrValueProvider, ValueProvider


class Filter(Step):
    """A `Filter` takes a given record and evaluates whether or not it should continue downstream.

    `Filter` steps generally make up the middle of an ETL pipeline and are responsible
    for ensuring only relevant records make it through.
    """

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        async for record in record_stream:
            if record is Flush or not await self.filter_record(record):
                yield record

    @abstractmethod
    async def filter_record(self, record: Any) -> bool:
        raise NotImplementedError


class ValueMatcher:
    @classmethod
    def from_file_data(
        cls,
        value: StaticValueOrValueProvider,
        possibilities: Iterable[StaticValueOrValueProvider],
        normalization: Optional[Dict[str, Any]] = None,
    ):
        return cls(
            value_provider=ValueProvider.guarantee_value_provider(value),
            possibilities=ValueProvider.guarantee_provider_list(possibilities),
            normalization=(normalization or {}),
        )

    def __init__(
        self,
        value_provider: ValueProvider,
        possibilities: Iterable[ValueProvider],
        normalization: Dict[str, Any],
    ) -> None:
        self.value_provider = value_provider
        self.possibilities = possibilities
        self.normalization = normalization

    def does_match(self, context: ProviderContext):
        actual_value = self.value_provider.normalize_single_value(
            context, self.normalization
        )

        return any(
            possibility_provider.normalize_single_value(context, self.normalization)
            == actual_value
            for possibility_provider in self.possibilities
        )


class ValuesMatchPossibilitiesFilter(Filter):
    """A filter that checks if a given value matches any of a set of possibilities."""

    @classmethod
    def from_file_data(cls, *, fields: Iterable[Dict[str, Any]]):
        value_matchers = [ValueMatcher.from_file_data(**field) for field in fields]
        return cls(value_matchers=value_matchers)

    def __init__(self, value_matchers: Iterable[ValueMatcher]):
        self.value_matchers = value_matchers

    async def filter_record(self, item):
        context_from_record = ProviderContext(item, None)
        return not all(
            matcher.does_match(context_from_record) for matcher in self.value_matchers
        )


class ExcludeWhenValuesMatchPossibilities(Filter):
    @classmethod
    def from_file_data(cls, **kwargs):
        inner = ValuesMatchPossibilitiesFilter.from_file_data(**kwargs)
        return cls(inner)

    def __init__(self, inner: ValuesMatchPossibilitiesFilter) -> None:
        self.inner = inner

    async def filter_record(self, record: Any) -> bool:
        return not await self.inner.filter_record(record)


class RegexMatcher:
    @classmethod
    def from_file_data(
        cls,
        value: StaticValueOrValueProvider,
        regex: str,
        include: bool = True,
        normalization: Optional[Dict[str, Any]] = None,
    ):
        return cls(
            value_provider=ValueProvider.guarantee_value_provider(value),
            regex=regex,
            include=include,
            normalization=(normalization or {}),
        )

    def __init__(
        self,
        value_provider: StaticValueOrValueProvider,
        regex: StaticValueOrValueProvider,
        include: bool,
        normalization: Dict[str, Any],
    ) -> None:
        self.value_provider = value_provider
        self.regex = re.compile(regex)
        self.include = include
        self.normalization = normalization

    def should_include(self, context: ProviderContext) -> bool:
        actual_value = self.value_provider.normalize_single_value(
            context, self.normalization
        )
        match = self.regex.match(actual_value) is not None
        return not match if self.include else match


class ValueMatchesRegexFilter(Filter):
    """A filter that includes/excludes a given value based on a matched regex."""

    @classmethod
    def from_file_data(
        cls,
        value: StaticValueOrValueProvider,
        regex: StaticValueOrValueProvider,
        include: bool,
        normalize: Optional[Dict[str, Any]] = None,
    ):
        regex_matcher = RegexMatcher.from_file_data(value, regex, include, normalize)
        return cls(regex_matcher=regex_matcher)

    def __init__(self, regex_matcher: RegexMatcher):
        self.regex_matcher = regex_matcher

    async def filter_record(self, item: Any):
        context_from_record = ProviderContext(item, None)
        return self.regex_matcher.should_include(context_from_record)



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from ..file_io import LazyLoadedArgument, LazyLoadedTagSafeLoader, LoadsFromYamlFile
from .argument_resolvers import set_config
from .class_loader import ClassLoader
from .normalizers import Normalizer
from .pipeline import Pipeline
from .scope_config import ScopeConfig
from .step import Step
from .value_providers import ValueProvider


class InvalidPipelineDefinitionError(ValueError):
    """Raised when a pipeline definition is invalid."""

    pass


class PipelineFileSafeLoader(LazyLoadedTagSafeLoader):
    """A YAML loader that can load pipeline files.""" ""

    was_configured = False

    @classmethod
    def configure(cls):
        if cls.was_configured:
            return

        for normalizer in Normalizer.all():
            normalizer.setup()
        for value_provider in ValueProvider.all():
            value_provider.install_yaml_tag(cls)

        cls.was_configured = True


@dataclass(slots=True)
class PipelineInitializationArguments:
    """Arguments used to initialize a pipeline from a file."""

    step_outbox_size: int = 1000
    annotations: Optional[List[str]] = None
    on_effective_configuration_resolved: Optional[Callable[[List[Dict]], None]] = None
    extra_steps: Optional[List[Step]] = None
    effecitve_config_values: Optional[ScopeConfig] = None

    @classmethod
    def for_introspection(cls):
        return cls(annotations=["introspection"])

    @classmethod
    def for_testing(cls):
        return cls(annotations=["test"])

    def initialize_from_file_data(self, file_data: Any):
        with set_config(self.effecitve_config_values):
            return Pipeline(
                steps=self.load_steps(ClassLoader(Step), file_data),
                step_outbox_size=self.step_outbox_size,
            )

    def load_steps(self, class_loader, file_data):
        effective = self.get_effective_configuration(file_data)
        if self.on_effective_configuration_resolved:
            self.on_effective_configuration_resolved(file_data)
        in_file = [self.load_step(class_loader, **step_data) for step_data in effective]
        return in_file + (self.extra_steps or [])

    def load_step(self, class_loader, implementation, arguments=None, factory=None):
        arguments = LazyLoadedArgument.resolve_if_needed(arguments or {})
        return class_loader.load_class(
            implementation=implementation, arguments=arguments, factory=factory
        )

    def get_effective_configuration(self, file_data):
        return [
            step_data for step_data in file_data if self.should_load_step(step_data)
        ]

    def should_load_step(self, step):
        return self.step_is_tagged_properly(step)

    def step_is_tagged_properly(self, step):
        # By default, we need to load all steps. Only filter a step out if:
        #   1. We have set annotations during initialization (e.g from the cli)
        #   2. The step is annotated
        #   3. There is not an intersection between the annotations from (1) and (2).
        #
        # This function also has a side effect of removing the annotations from the step data.
        annotations_set_on_step = set(step.pop("annotations", []))
        if annotations_set_on_step and self.annotations:
            return annotations_set_on_step.intersection(self.annotations)

        return True


class PipelineFileContents(LoadsFromYamlFile):
    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            [
                {
                    "implementation": str,
                    Optional("factory"): str,
                    Optional("annotations"): [str],
                    Optional("arguments"): {str: object},
                }
            ]
        )

    @classmethod
    def get_loader(cls):
        PipelineFileSafeLoader.configure()
        return PipelineFileSafeLoader

    @classmethod
    def from_file_data(cls, data):
        return cls(data)

    def __init__(self, data: List[Dict]) -> None:
        self.data = data

    def initalize_with_arguments(self, init_args: PipelineInitializationArguments):
        return init_args.initialize_from_file_data(self.data)


class PipelineFile:
    def __init__(self, file_path: Path):
        self.file_path = file_path

    def load_pipeline(
        self, init_args: Optional[PipelineInitializationArguments] = None
    ) -> Pipeline:
        init_args = init_args or PipelineInitializationArguments()
        contents = PipelineFileContents.read_from_file(self.file_path)
        return contents.initalize_with_arguments(init_args)



from abc import ABC, abstractclassmethod, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Type

from schema import Schema
from yaml import SafeDumper, SafeLoader, dump, load


class DescribesYamlSchema(ABC):
    """A mixin for classes that can be described by a YAML schema."""

    @abstractclassmethod
    def describe_yaml_schema(self) -> Schema:
        raise NotImplementedError


class LoadsFromYaml(DescribesYamlSchema):
    """A mixin for classes that can be read from YAML."""

    @abstractclassmethod
    def from_file_data(cls, data):
        """Create an instance of this class from the given file data.

        Args:
            data: The data read from the file.

        Returns:
            An instance of this class.
        """
        raise NotImplementedError

    @classmethod
    def validate_and_load(cls, data):
        """Validate the given data against the YAML schema and load it into an instance of this class.

        Args:
            data: The data to validate and load.

        Raises:
            SchemaError: If the data does not match the YAML schema.

        Returns:
            An instance of this class.
        """
        validated_data = cls.describe_yaml_schema().validate(data)
        return cls.from_file_data(validated_data)


class LoadsFromYamlFile(LoadsFromYaml):
    """A mixin for classes that can be read from a YAML file."""

    @classmethod
    def get_loader(cls) -> Type[SafeLoader]:
        """Get the YAML loader to use when reading this object from a file.

        Returns:
            The YAML loader to use when reading this object from a file.
        """
        return SafeLoader

    @classmethod
    def read_from_file(cls, file_path: Path) -> "LoadsFromYaml":
        """Read this object from a YAML file.

        Args:
            file_path: The path to the file to read from.

        Raises:
            FileNotFoundError: If the file does not exist.

        Returns:
            An instance of this class.
        """
        if not file_path.exists():
            raise FileNotFoundError(f"File '{file_path}' does not exist.")

        with file_path.open("r") as f:
            file_data = load(f, Loader=cls.get_loader())
            return cls.validate_and_load(file_data)


class SavesToYaml(DescribesYamlSchema):
    """A mixin for classes that can be written to as YAML."""

    @abstractmethod
    def to_file_data(self):
        """Get the data to write to a YAML file.

        Returns:
            The data to write to a YAML file.
        """
        raise NotImplementedError


class SavesToYamlFile(SavesToYaml):
    """A mixin for classes that can be written to a YAML file."""

    @classmethod
    def get_dumper(cls) -> Type[SafeDumper]:
        """Get the YAML dumper to use when writing this object to a file.

        Returns:
            The YAML dumper to use when writing this object to a file.
        """
        return SafeDumper

    def write_to_file(self, file_path: Path):
        """Write this object to a YAML file.

        If the file already exists, it will be overwritten.

        Args:
            file_path: The path to the file to write to.

        Returns:
            None
        """
        file_data = self.to_file_data()
        validated_file_data = self.describe_yaml_schema().validate(file_data)
        with file_path.open("w") as f:
            dump(
                validated_file_data,
                f,
                Dumper=self.get_dumper(),
                indent=2,
                sort_keys=True,
            )


# This approach is inspired by https://death.andgravity.com/any-yaml
#
# Generally, the idea is that instead of trying to resolve the arguments at the time of parsing the yaml file,
# we instead create a special object that will resolve the arguments at the time of use.
#
# This comes with the side effect that we shift the responsibility of resolving the arguments from the
# yaml parser to the code that uses the "loaded" yaml data.
@dataclass(frozen=True, slots=True)
class LazyLoadedArgument:
    tag: str
    value: Any

    def get_value(self):
        from .pipeline.argument_resolvers import ArgumentResolver

        return ArgumentResolver.resolve_argument_with_alias(self.tag, self.value)

    @staticmethod
    def resolve_if_needed(value):
        if isinstance(value, LazyLoadedArgument):
            return value.get_value()
        if isinstance(value, dict):
            return {
                k: LazyLoadedArgument.resolve_if_needed(v) for k, v in value.items()
            }
        if isinstance(value, list):
            return [LazyLoadedArgument.resolve_if_needed(v) for v in value]
        return value


class LazyLoadedTagSafeLoader(SafeLoader):
    pass


def wrap_unloaded_tag(self, node):
    value = self.construct_scalar(node)
    return LazyLoadedArgument(node.tag[1:], value)


LazyLoadedTagSafeLoader.add_constructor(None, wrap_unloaded_tag)
SafeDumper.add_representer(
    LazyLoadedArgument,
    lambda dumper, arg: dumper.represent_scalar(arg.tag, arg.value),
)



import logging
import os
from typing import Any

from pythonjsonlogger.jsonlogger import JsonFormatter

from ...pipeline.meta import get_context
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


def _get_logger_level():
    return logging.getLevelName(os.environ.get("NODESTREAM_LOG_LEVEL", "INFO").upper())


def configure_logging_with_json_defaults():
    logging.basicConfig(level=_get_logger_level(), force=True)
    old_record_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record = old_record_factory(*args, **kwargs)
        record.pipeline_name = get_context().name
        return record

    logging.setLogRecordFactory(record_factory)

    formatter = JsonFormatter(
        "%(name)s %(levelname)s %(pipeline_name)s %(message)s", timestamp=True
    )
    logger = logging.getLogger()  # Configure the root logger.
    logger.handlers[0].setFormatter(formatter)


class InitializeLogger(Operation):
    async def perform(self, command: NodestreamCommand) -> Any:
        if command.has_json_logging_set:
            configure_logging_with_json_defaults()



from typing import List

from ...databases import Copier, GraphDatabaseWriter
from ...pipeline import Pipeline
from ...project import Project, Target
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class RunCopy(Operation):
    def __init__(
        self,
        from_target: Target,
        to_target: Target,
        project: Project,
        node_types: List[str],
        relationship_types: List[str],
    ) -> None:
        self.from_target = from_target
        self.to_target = to_target
        self.project = project
        self.node_types = node_types
        self.relationship_types = relationship_types

    async def perform(self, command: NodestreamCommand):
        pipeline = self.build_pipeline()
        await pipeline.run()

    def build_pipeline(self) -> Pipeline:
        copier = self.build_copier()
        writer = self.build_writer()
        return Pipeline([copier, writer], step_outbox_size=10000)

    def build_copier(self) -> Copier:
        return Copier(
            self.from_target.make_type_retriever(),
            self.project,
            self.node_types,
            self.relationship_types,
        )

    def build_writer(self) -> GraphDatabaseWriter:
        return self.to_target.make_writer()



from .graph_schema_extraction import LargeLanguageModelSchemaPrinter
from .graphql_schema_printer import GraphQLSchemaPrinter
from .plain_text_schema_printer import PlainTestSchemaPrinter
from .schema_printer import SCHEMA_PRINTER_SUBCLASS_REGISTRY, SchemaPrinter

__all__ = (
    "SchemaPrinter",
    "GraphQLSchemaPrinter",
    "LargeLanguageModelSchemaPrinter",
    "PlainTestSchemaPrinter",
    "SCHEMA_PRINTER_SUBCLASS_REGISTRY",
)



import asyncio
import importlib.metadata
import os
import sys

from cleo.application import Application

from ..project.audits import Audit
from .commands import AuditCommand, NodestreamCommand

PACKAGE_NAME = "nodestream"


def get_version() -> str:
    return importlib.metadata.version(PACKAGE_NAME)


def get_application() -> Application:
    app = Application(PACKAGE_NAME, get_version())
    for audit in Audit.all():
        app.add(AuditCommand.for_audit(audit)())
    for command in NodestreamCommand.all():
        if command not in (AuditCommand, NodestreamCommand):
            app.add(command())
    return app


def run():
    sys.path.append(os.getcwd())

    try:
        import uvloop

        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    except ImportError:
        pass

    get_application().run()



from .pipeline_definition import PipelineConfiguration, PipelineDefinition
from .pipeline_scope import PipelineScope
from .project import Project, ProjectPlugin
from .run_request import RunRequest
from .target import Target

__all__ = (
    "PipelineDefinition",
    "PipelineConfiguration",
    "PipelineScope",
    "Project",
    "ProjectPlugin",
    "RunRequest",
    "Target",
)



from .copy import Copier, TypeRetriever
from .database_connector import DatabaseConnector
from .debounced_ingest_strategy import DebouncedIngestStrategy
from .writer import GraphDatabaseWriter

__all__ = (
    "GraphDatabaseWriter",
    "DebouncedIngestStrategy",
    "DatabaseConnector",
    "TypeRetriever",
    "Copier",
)



from typing import Any

from .normalizer import Normalizer


class RemoveTrailingDots(Normalizer, alias="remove_trailing_dots"):
    def normalize_value(self, value: Any) -> Any:
        return value.rstrip(".") if isinstance(value, str) else value



from .expand_json_field import ExpandJsonField
from .transformer import (
    ConcurrentTransformer,
    PassTransformer,
    SwitchTransformer,
    Transformer,
)
from .value_projection import ValueProjection

__all__ = (
    "ExpandJsonField",
    "ValueProjection",
    "Transformer",
    "ConcurrentTransformer",
    "SwitchTransformer",
    "PassTransformer",
)



from typing import Optional
from uuid import uuid4

import boto3
from botocore.credentials import RefreshableCredentials
from botocore.session import Session


class AwsClientFactory:
    def __init__(
        self,
        assume_role_arn: Optional[str] = None,
        assume_role_external_id: Optional[str] = None,
        **boto_session_args
    ) -> None:
        self.assume_role_arn = assume_role_arn
        self.assume_role_external_id = assume_role_external_id
        self.session_args = self._init_session_args(**boto_session_args)

    @staticmethod
    def _init_session_args(**boto_session_args):
        session_args = boto_session_args.copy() if boto_session_args else {}
        if "profile_name" in session_args and not session_args["profile_name"]:
            del session_args["profile_name"]
        return session_args

    def assume_role_and_get_credentials(self):
        sts_client = boto3.client("sts", **self.session_args)
        role_session_name = str(uuid4())
        assume_role_kwargs = {
            "RoleArn": self.assume_role_arn,
            "RoleSessionName": role_session_name,
        }
        if self.assume_role_external_id:
            assume_role_kwargs["ExternalId"] = self.assume_role_external_id
        assumed_role_object = sts_client.assume_role(**assume_role_kwargs)
        return {
            "access_key": assumed_role_object["Credentials"]["AccessKeyId"],
            "secret_key": assumed_role_object["Credentials"]["SecretAccessKey"],
            "token": assumed_role_object["Credentials"]["SessionToken"],
            "expiry_time": assumed_role_object["Credentials"]["Expiration"].isoformat(),
        }

    def get_boto_session_with_refreshable_credentials(self):
        refreshable_credentials = RefreshableCredentials.create_from_metadata(
            metadata=self.assume_role_and_get_credentials(),
            refresh_using=self.assume_role_and_get_credentials,
            method="sts-assume-role",
        )
        session = Session()
        session._credentials = refreshable_credentials
        return session

    def assume_role_if_supplied_and_get_session(self):
        if self.assume_role_arn:
            return self.get_boto_session_with_refreshable_credentials()
        return Session()

    def make_client(self, client_name: str):
        session = self.assume_role_if_supplied_and_get_session()
        return boto3.Session(botocore_session=session, **self.session_args).client(
            client_name
        )



from abc import ABC, abstractmethod
from typing import AsyncGenerator, List

from ..model import Node, RelationshipWithNodes
from ..pipeline import Extractor
from ..project import Project
from ..schema.indexes import KeyIndex


class TypeRetriever(ABC):
    @abstractmethod
    def get_nodes_of_type(self, shape: str) -> AsyncGenerator[Node, None]:
        raise NotImplementedError

    @abstractmethod
    def get_relationships_of_type(
        self, type: str
    ) -> AsyncGenerator[RelationshipWithNodes, None]:
        raise NotImplementedError


class Copier(Extractor):
    def __init__(
        self,
        type_retriver: TypeRetriever,
        project: Project,
        node_types: List[str],
        relationship_types: List[str],
    ) -> None:
        self.type_retriever = type_retriver
        self.relationship_types = relationship_types
        self.node_types = node_types
        self.keys_by_node_type = {
            key.type: key
            for key in project.gather_used_indexes()
            if isinstance(key, KeyIndex)
        }

    async def extract_records(self):
        for node_type in self.node_types:
            nodes = self.type_retriever.get_nodes_of_type(node_type)
            async for node in nodes:
                yield self.convert_node_to_ingest(node)

        for rel_type in self.relationship_types:
            rels = self.type_retriever.get_relationships_of_type(rel_type)
            async for relationship in rels:
                yield self.convert_relationship_to_ingest(relationship)

    def reorganize_node_key_properties(self, node: Node):
        # This is a bit of a hack, but it's the only way to make sure that the
        # keys are in the right place. We need to remove the keys from the
        # properties and put them in the key_values because otherwise
        # we have to push a lot of work down to the database connector to ensure
        # that the keys are in the right place. Its easier to just do it here.
        key = self.keys_by_node_type[node.type]
        for key_name in key.identity_keys:
            node.key_values[key_name] = node.properties[key_name]
            del node.properties[key_name]

    def convert_node_to_ingest(self, node: Node):
        self.reorganize_node_key_properties(node)
        return node.into_ingest()

    def convert_relationship_to_ingest(self, relationship: RelationshipWithNodes):
        self.reorganize_node_key_properties(relationship.from_node)
        self.reorganize_node_key_properties(relationship.to_node)
        return relationship.into_ingest()



import json
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Any, Iterable

from ....model import JsonLikeDocument
from ....pluggable import Pluggable
from ....subclass_registry import SubclassRegistry
from ...flush import Flush
from ..extractor import Extractor

STREAM_CONNECTOR_SUBCLASS_REGISTRY = SubclassRegistry()
STREAM_OBJECT_FORMAT_SUBCLASS_REGISTRY = SubclassRegistry()


@STREAM_CONNECTOR_SUBCLASS_REGISTRY.connect_baseclass
class StreamConnector(Pluggable, ABC):
    entrypoint_name = "stream_connectors"

    @abstractmethod
    async def connect(self):
        raise NotImplementedError

    @abstractmethod
    async def disconnect(self):
        raise NotImplementedError

    @abstractmethod
    async def poll(self) -> Iterable[Any]:
        raise NotImplementedError


@STREAM_OBJECT_FORMAT_SUBCLASS_REGISTRY.connect_baseclass
class StreamRecordFormat(Pluggable, ABC):
    entrypoint_name = "record_formats"

    @abstractmethod
    def parse(self, record: Any) -> JsonLikeDocument:
        raise NotImplementedError


class JsonStreamRecordFormat(StreamRecordFormat, alias="json"):
    def parse(self, record: Any) -> JsonLikeDocument:
        return json.loads(record)


class StreamExtractor(Extractor):
    """A StreamExtractor implements the standard behavior of polling data from a stream.

    The StreamExtractor requires both a StreamConnector and a StreamRecordFormat to delegate
    to for the actual polling implementation and parsing of the data, respectively.
    """

    @classmethod
    def from_file_data(cls, connector: str, record_format: str, **connector_args):
        # Import all plugins so that they can register themselves
        StreamRecordFormat.import_all()
        StreamConnector.import_all()

        object_format_cls = STREAM_OBJECT_FORMAT_SUBCLASS_REGISTRY.get(record_format)
        connector_cls = STREAM_CONNECTOR_SUBCLASS_REGISTRY.get(connector)
        return cls(
            record_format=object_format_cls(),
            connector=connector_cls(**connector_args),
        )

    def __init__(
        self,
        connector: StreamConnector,
        record_format: StreamRecordFormat,
    ):
        self.connector = connector
        self.record_format = record_format
        self.log_flush = True
        self.logger = getLogger(__name__)

    async def poll(self):
        results = await self.connector.poll()
        if len(results) == 0:
            if self.log_flush:
                self.logger.info("flushing extractor because no records were returned")
                self.log_flush = False
            yield Flush
        else:
            self.logger.debug("Received Kafka Messages")
            self.log_flush = True
            for record in results:
                yield self.record_format.parse(record)

    async def extract_records(self):
        await self.connector.connect()
        try:
            while True:
                async for item in self.poll():
                    yield item
        except Exception:
            self.logger.exception("failed extracting records")
        finally:
            await self.connector.disconnect()



import re
from typing import Any

from ..commands.nodestream_command import NodestreamCommand


class Operation:
    @property
    def name(self) -> str:
        class_name = self.__class__.__name__
        return " ".join(re.findall(r"[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))", class_name))

    async def perform(self, command: NodestreamCommand) -> Any:
        raise NotImplementedError



from cleo.helpers import argument, option

PROJECT_FILE_OPTION = option(
    "project", "p", "The project file (nodestream.yaml) to load.", flag=False
)

DATABASE_NAME_OPTION = option(
    "database", "d", "The Database to Configure", flag=False, default="neo4j"
)

SCOPE_NAME_OPTION = option(
    "scope", "s", "Use the provided scope for the operation", flag=False
)

JSON_OPTION = option("json", "j", "Log output in JSON", flag=True)


PIPELINE_ARGUMENT = argument("pipeline", "the name of the pipeline")
MANY_PIPELINES_ARGUMENT = argument(
    "pipelines", "the names of the pipelines", multiple=True, optional=True
)



from abc import ABC, abstractmethod

from ...pipeline.value_providers import ProviderContext
from ...pluggable import Pluggable
from ...schema.schema import IntrospectiveIngestionComponent
from ...subclass_registry import SubclassRegistry

INTERPRETATION_REGISTRY = SubclassRegistry()


@INTERPRETATION_REGISTRY.connect_baseclass
class Interpretation(IntrospectiveIngestionComponent, Pluggable, ABC):
    entrypoint_name = "interpretations"

    @abstractmethod
    def interpret(self, context: ProviderContext):
        raise NotImplementedError

    @classmethod
    def from_file_data(cls, **arguments) -> "Interpretation":
        name = arguments.pop("type")
        class_to_load = INTERPRETATION_REGISTRY.get(name)
        return class_to_load(**arguments)

    def gather_used_indexes(self):
        yield from []

    def gather_object_shapes(self):
        yield from []

    def gather_present_relationships(self):
        yield from []



from .class_loader import ClassLoader
from .extractors import Extractor, IterableExtractor
from .filters import (
    ExcludeWhenValuesMatchPossibilities,
    Filter,
    ValueMatchesRegexFilter,
    ValuesMatchPossibilitiesFilter,
)
from .flush import Flush
from .meta import UNKNOWN_PIPELINE_NAME
from .pipeline import Pipeline
from .pipeline_file_loader import PipelineFile, PipelineInitializationArguments
from .progress_reporter import PipelineProgressReporter
from .scope_config import ScopeConfig
from .step import PassStep, Step
from .transformers import Transformer
from .writers import LoggerWriter, Writer

__all__ = (
    "ClassLoader",
    "Extractor",
    "IterableExtractor",
    "Filter",
    "ValuesMatchPossibilitiesFilter",
    "ValueMatchesRegexFilter",
    "ExcludeWhenValuesMatchPossibilities",
    "PipelineFile",
    "Pipeline",
    "ScopeConfig",
    "Step",
    "Transformer",
    "Writer",
    "LoggerWriter",
    "PipelineInitializationArguments",
    "PassStep",
    "Flush",
    "UNKNOWN_PIPELINE_NAME",
    "PipelineProgressReporter",
)



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Optional

from ..scope_config import ScopeConfig
from .argument_resolver import ArgumentResolver

config: ContextVar[Optional[ScopeConfig]] = ContextVar("context")


def get_config() -> Optional[ScopeConfig]:
    try:
        return config.get()
    except LookupError:
        return None


@contextmanager
def set_config(new_config: Optional[ScopeConfig]):
    token = config.set(new_config)
    try:
        yield
    finally:
        config.reset(token)


class ConfigurationArgumentResolver(ArgumentResolver, alias="config"):
    """An `ConfigurationArgumentResolver` is an `ArgumentResolver` that can resolves configuration values."""

    @staticmethod
    def resolve_argument(variable_name):
        if config := get_config():
            return config.get_config_value(variable_name)
        else:
            return None
